var fs = require("fs");
var http = require("http");
var sys = require("sys");
var url = require("url");
var util = require("util");

var _ = require("./underscore")._;
var json = require("./json2");

// var Client = require("mysql").Client;
var sqlite = require("sqlite");

var result = require("./result").result;

var silo = "silo/";

var tables = {};
var databases = {};

var csf = function () {
    res.writeHead(200, {'Content-Type': 'text/plain'});
    res.end('Hello World\n');
};

var started = function () {
    util.log('Server running at http://127.0.0.1:33114/');
};

var request_handler = function (request, response) {
    var method = request.method;
    var u = request.url;
    var parsed = url.parse(u, true);
    var path = parsed.pathname.slice(1);
    var query = parsed.query;
    var first = path.split('/')[0];
    
    if (method === "GET") {
        switch (first) {
        case "get":
            get(request, response, path, query);
            break;
        case "del":
            del(request, response, path, query);
            break;
        case "delview":
        case "getview":
        case "tableopts":
            unimplemented(request, response, path, query);
            break;
        default:
            unknown(request, response, path, query);
        }
    } else if (method === "POST") {
        switch (first) {
        case "create":
            create(request, response, path, query);
            break;
        case "put":
            put(request, response, path, query);
            break;
        case "putview":
        case "safeput":
        case "modtableopts":
            unimplemented(request, response, path, query);
        default:
            unknown(request, response, path, query);
        }
    } else {
        // ?
        unknown(request, response, path, query);
    }
};

var get = function (request, response, path, query) {
    var splat = path.split("/");
    var database_name = splat[1];
    var key = splat[2];
    var db = databases[database_name];
    
    util.log("i was asked about " + database_name + ":" + key);
    
    if (db === undefined) {
        var why = "unknown database: " + database_name;
        response.writeHead(200, {"Content-Type": "application/json"});
        response.end(json.stringify(result("failed", why)));
        return;
    }
    
    var sql_query = "SELECT value FROM data WHERE key = '" + key + "';";
    db.execute(sql_query, [], function (error, rows) {
        if (error) {
            // send an error back to client
            throw error;
        }

        sys.debug("got back " + rows.length + " rows.");
        if (rows.length != 1) {
            if (rows.length === 0) {
                // nothing to get
                response.writeHead(200, {"Content-Type": "application/json"});
                response.end(json.stringify(result("failed", "key not found")));
            } else {
                // something else went wrong; send msg back to client
                response.writeHead(200, {"Content-Type": "application/json"});
                response.end(json.stringify(result("failed", "internal error")));
            }
            return;
        }

        var value0 = rows[0].value;
        try {
            var value = json.parse(value0);
        } catch (e) {
            response.writeHead(200, {"Content-Type": "application/json"});
            response.end(json.stringify(result("failed", "invalid value stored in database (key: " + key + "; value stored: " + value0 + ")")));
        }
        response.writeHead(200, {"Content-Type": "application/json"});
        response.end(json.stringify(result("ok", value)));
    });
};

var del = function (request, response, path, query) {
    var splat = path.split("/");
    var database_name = splat[1];
    var key = splat[2];
    var db = databases[database_name];
    
    if (db === undefined) {
        var why = "unknown database: " + database_name;
        response.writeHead(200, {"Content-Type": "application/json"});
        response.end(json.stringify(result("failed", why)));
        return;
    }

    var sql_query = "DELETE FROM data WHERE key = '" + key + "';";
    db.execute(sql_query, [], function (error, rows) {
        if (error) {
            throw error;
        }

        // otherwise everything went ok?
        response.writeHead(200, {"Content-Type": "application/json"});
        response.end(json.stringify(result("ok", null)));
    });
};

var put = function (request, response, path, query) {
    var splat = path.split("/");
    var database_name = splat[1];
    var key = splat[2];
    var db = databases[database_name];
    
    var buf = "";

    sys.debug("in put");
    
    request.setEncoding("utf8");
    
    request.on('data', function (chunk) {
        util.debug("request.on#data: got: " + chunk);
        buf += chunk;
    });
        
    request.on('end', function () {
        try {
            // For now, we throw the parsed value away.  We only check whether we've
            // been given valid json or not.  Later, we can use it to create views
            // and maybe other interesting things as well.
            var parsed = json.parse(buf);
        } catch (e) {
            util.log("got some bogus json");
            response.writeHead(200, {"Content-Type": "application/json"});
            response.end(json.stringify(result("failed", "invalid json")));
            return;
        }
        
        // This is, uh, not really safe.  I refer you back to README.md:
        // "I wouldn't use mynosqlitegres in production."
        var sql_query = "INSERT INTO data VALUES ('" + key + "', '" + buf + "');";
        
        db.execute(sql_query, function (error, rows) {
            if (error) {
                throw error;
            }

            _.each(rows, function (row) {
                util.debug("request.on#db.execute#callbackgot: " + util.inspect(row));
            });
            
            response.writeHead(200, {"Content-Type": "application/json"});
            response.end(json.stringify(result("ok", null)));
        });
    });
};

var create = function (request, response, path, query) {
    var splat = path.split("/");
    var database_name = splat[1];
    var db = new sqlite.Database();

    var schema = "CREATE TABLE data (key VARCHAR, value VARCHAR, PRIMARY KEY (key);";

    var file = silo + "/" + database_name;

    if (databases[database_name] !== null) {
        util.debug("not creating database " + database_name +
                   " because it already exists");
        response.writeHead(200, {"Content-Type": "application/json"});
        response.end(json.stringify(result("failed", database_name + " already exists")));
    }
    
    db.open(file, function (error) {
        if (error) {
            response.writeHead(200, {"Content-Type": "application/json"});
            response.end(json.stringify(result("failed", "error opening " + file + ":" + error)));
            return;
        }
        
        db.execute(schema, [], function (error, rows) {
            if (error) {
                response.writeHead(200, {"Content-Type": "application/json"});
                response.end(json.stringify(result("failed", "failed to initialize database:" + error )));
                return;
                
            }
            response.writeHead(200, {"Content-Type": "application/json"});
            response.end(json.stringify(result("ok", null)));
        });
    });
};

var unimplemented = function(request, response, path, query) {
    var first = path.split("/")[0];
    generic_error(response, "unimplemented method: " + first);
};

var unknown = function (request, response, path, query) {
    var first = path.split("/")[0];
    generic_error(response, "unknown method: " + first);
};

var generic_error = function (response, text) {
    response.writeHead(200, {"Content-Type": "application/json"});
    response.end(json.stringify(result("failed", text)));
};

var post_handler = function (request, response) {
    
};

// Fake syncyonously loading already-existing databases.
var initial_load = function () {
    var dbs0 = fs.readdirSync(silo);
    
    var done = 0;
    
    util.log("performing initial load");
    
    var dbs = _.filter(dbs0, function (filename) {
        if (filename !== ".gitignore")
            return true;
        return false;
    });
    var goal = dbs.length;
    
    for (var i = 0; i < dbs.length; ++i) {
        var db = new sqlite.Database();
        var database = dbs[i];
        
        var results = [];
        var file = silo + "/" + database;

        db.open(file, (function (d) {
            util.log("opening database file: " + file);
            return function() {
                util.log("loading database: " + database);

                var sql = "SELECT COUNT(*) FROM data;";
                db.execute(sql, [], function (error, rows) {
                    if (error) {
                        util.debug("problem opening " + file + ": " + error);
                        throw error;
                    }
                    
                    done++;
                    util.log("database " + file + ": loaded");
                    if (done === goal) {
                        util.log("all databases loaded");
                        go();
                    }
                });
            };
        })(db));
        
        databases[database] = db;
    }
};

var go = function () {
    var server = http.createServer(/*csf*/);
    
    server.on('request', request_handler);
    
    server.listen(33114, "127.0.0.1", started);
};

initial_load();
